#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include "const.h"
#include "types.h"
#include "random.h"

/* Play around with those constants */
#define ALPHA 2.0
#define MU -2.2
#define OMEGA_P 1.0
#define TMAX 50.0
#define NBINS 1000
#define STATISTIC_FREQUENCY 10000000
#define TEST_FREQUENCY 10000000
#define CHECK_TIME_FREQUENCY 100000000
#define RUNTIME 3600
#define LOWEST_ORDER 0
#define HIGHEST_ORDER 200

/* Turn initial tests on or off - only used in non-productive runs */

const double twopi = 2.0 * PI;
const double vsqu = 4.0 * ALPHA * PI / SQRT_TWO;

/*
	ABOUT LINUX COLOR ESCAPE CODES:
	http://www.linuxjournal.com/article/8603

	LIST OF COLORS:
	http://www.linuxforums.org/forum/programming-scripting/88-color-console.html

*/

// Current order of the diagram
int order;
// Number of rejects
ulong rejects;
// Bin size
double dt;
// Current diagram time
double t;
// Current diagram index
int tindex;
// Initial momentum
vec3 k;
// Diagram nodes
node **nodes;
// Statistic histograms
double *G;
// Statistic hits per order
int *Z;
// Norm of the diagram
double norm;

/*

	Calculation helpers

*/

// Computes the energy given by the absolute momentum
double ScalarEnergy(double p)
{
	return p * p / 2.0 - MU;
}

// Computes the energy given by the momentum vector
double VectorEnergy(vec3 *p)
{
	return SquareVec3(p) / 2.0 - MU;
}

// Computes the electron propagator given by the absolute momentum and the time difference
double ScalarPropagator(double p, double dt)
{
	return exp(-ScalarEnergy(p) * dt);
}

// Computes the electron propagator given by the momentum vector and the time difference
double VectorPropagator(vec3 *p, double dt)
{
	return exp(-VectorEnergy(p) * dt);
}

// Computes the phonon propagator given by the time difference
double Phonon(double dt)
{
	return exp(-OMEGA_P * dt);
}

/*

	Statistic

*/

// Performs the measurement on the current diagram
void Measure()
{
	G[tindex] = G[tindex] + 1.0;
	Z[order]++;
}

// Computes the green function at a given time t with the current statistics
double Green(double t)
{
	int i = (int)floor(t / dt);
    return G[i] / dt * norm / Z[0];
}

// Prints the current statistics
void PrintStatistic(ulong steps)
{
	int i;

	printf("Accepted diagram changes    =  %lu\n", steps - rejects);
	printf("Rejected diagram changes    =  %lu%%\n", rejects * 100 / steps);

	for(i = 0; i <= HIGHEST_ORDER; i++)
	{
		printf("Normalization statistics %d  =  %d\n", i, Z[i]);
	}

	printf("------------------------------------------\n");
}

// Saves the current statistics
void SaveStatistic()
{
	int i;
	double t = dt / 2.0;
	FILE *file = fopen("statistic.dat", "w");
	char* line = (char*)malloc(sizeof(char) * STRING_BUFFER);
	sprintf(line, "%s\t%s\t%s\n", "t", "g", "log(g)");
	fputs(line, file);

	for(i = 0; i < NBINS; i++)
	{
		double g = Green(t);
		line = (char*)malloc(sizeof(char) * STRING_BUFFER);
		sprintf(line, "%lf\t%lf\t%lf\n", t, g, log(g));
		fputs(line, file);
		t = t + dt;
	}

	fclose(file);
}

/*

	Arcs

*/

// Gets the number of non-crossing arcs
int GetNumberOfArcs()
{
	int i;
	int arcs = 0;
	int n = 2 * order;

	for(i = 1; i < n; i++)
	{
		if(nodes[i]->link == nodes[i - 1])
		{
			arcs++;
		}
	}

	return arcs;
}

// Gets the index (in the nodes-array) of a given arc (index)
int GetIndexOfArc(int arc)
{
	int i;
	int index = -1;
	int arcs = 0;
	int n = 2 * order;

	for(i = 1; i < n; i++)
	{
		if(nodes[i]->link == nodes[i - 1])
		{
			if(arc == arcs)
			{
				index = i - 1;
				break;
			}

			arcs++;
		}
	}

	return index;
}

/*

	Insert and delete (pairs of) nodes

*/

// Deletes a pair of nodes given by the position of the first node
void DeleteNodes(int position)
{
	int i;
	int n = 2 * order - 2;

	free(nodes[position + 1]);
	free(nodes[position]);

	for(i = position; i < n; i++)
	{
		nodes[i] = nodes[i + 2];
	}

	order--;
}

// Inserts a pair of nodes given by the insertation position (first node) and its data
void InsertNodes(int position, vec3 p, vec3 q, double ti, double tf)
{
	int i;
	node* initial;
	node* final;

	for(i = 2 * order - 1; i >= position; i--)
	{
		nodes[i + 2] = nodes[i];
	}

	initial = (node*)malloc(sizeof(node));
	final = (node*)malloc(sizeof(node));
	initial->t = ti;
	final->t = tf;
	initial->link = final;
	final->link = initial;
	initial->pin = p;
	final->pout = p;
	initial->q = q;
	final->q = q;
	SubtractVec3(&p, &q, &(initial->pout));
	final->pin = initial->pout;
	nodes[position] = initial;
	nodes[position + 1] = final;
	order++;
}

/*

	Updates

*/

// Tries to swap two non-connected nodes, i.e. rearranges 4 nodes (2 arcs)
void SwapDiagrams()
{
	if(order < 2)
	{
		rejects++;
	}
	else
	{
		int n = 2 * order - 1;
		int index = Discrete(0, n);
		node* first = nodes[index];
		node* second = nodes[index + 1];

		if(first->link == second)
		{
			rejects++;
		}
		else
		{
			double prob;
			double chance;
			double dt = second->t - first->t;
			node* firstLink = first->link;
			node* secondLink = second->link;
			vec3 oldMomentum = first->pout;
			vec3 newMomentum;

#ifdef DEBUG
			printf("index = %d\n", index);
#endif

			if(secondLink->t > second->t)
			{
				SubtractVec3(&(first->pin), &(second->q), &newMomentum);
			} 
			else
			{
				AddVec3(&(first->pin), &(second->q), &newMomentum);
			}

			double oldDelta = abs(first->t - firstLink->t) + abs(second->t - secondLink->t);
			double newDelta = abs(first->t - secondLink->t) + abs(second->t - firstLink->t);
			chance = VectorPropagator(&newMomentum, dt) * Phonon(newDelta) / (VectorPropagator(&oldMomentum, dt) * Phonon(oldDelta));
			prob = Random();

#ifdef DEBUG
			printf("oldDelta = %lf, newDelta = %lf, |q_old| = %lf\n", oldDelta, newDelta, SquareVec3(&oldMomentum));
			printf("|q_new| = %lf, R = %lf, Prob. = %lf\n", SquareVec3(&newMomentum), chance, prob);
#endif

			if(prob < chance)
			{
				first->link = secondLink;
				second->link = firstLink;
				firstLink->link = second;
				secondLink->link = first;

				first->pout = newMomentum;
				second->pin = newMomentum;

				first->q = secondLink->q;
				second->q = firstLink->q;
			}
			else
			{
				rejects++;
			}
		}
	}
}

// Tries to remove a diagram
void RemoveDiagram()
{
	if(order < 1)
	{
		rejects++;
	}
	else
	{
		double chance;
		double prob;
		int n = 2 * order - 1;
		int narcs = GetNumberOfArcs();

		if(narcs == 0)
		{
			rejects++;
		}
		else 
		{
			int arc = Discrete(0, narcs);
			int index = GetIndexOfArc(arc);
			double tauMin = 0.0;
			double tauMax = t;
			double dtau;
			double ti = nodes[index]->t;
			double tf = nodes[index + 1]->t;
			double dt = tf - ti;
			vec3 q = nodes[index]->q;
			vec3 p = nodes[index]->pin;

#ifdef DEBUG
			printf("index = %d, narcs = %d, n = %d\n", index, narcs, n);
#endif

			if (index > 0)
			{
				tauMin = nodes[index - 1]->t;
			}

			if (index + 1 < n)
			{
				tauMax = nodes[index + 2]->t;
			}

			dtau = tauMax - tauMin;
			prob = Random();
			chance = OMEGA_P * SquareVec3(&q) * narcs / (pow(1.0 / sqrt(twopi * dt), 3.0) * vsqu * n * dtau * exp(dt * DotMultiplayVec3(&p, &q)));

#ifdef DEBUG
			printf("ti = %lf, tf = %lf, dtau = %lf\n", ti, tf, dtau);
			printf("|q| = %lf, R = %lf, Prob. = %lf\n", SquareVec3(&q), chance, prob);
#endif

			if(prob < chance)
			{
				DeleteNodes(index);
			}
			else
			{
				rejects++;
			}
		}
	}
}

// Tries to add a diagram
void AddDiagram()
{
	if(order == HIGHEST_ORDER)
	{
		rejects++;
	}
	else
	{
		vec3 q;
		vec3 p = k;
		double sigma;
		double ti;
		double tf;
		double dt;
		double tauMin = 0.0;
		double tauMax = t;
		double dtau;
		double chance;
		double prob;
		int n = 2 * order + 1;
		int index = Discrete(0, n);
		int narcs = GetNumberOfArcs();

#ifdef DEBUG
		printf("index = %d, narcs = %d, n = %d\n", index, narcs, n);
#endif

		if(index == 0)
		{
			tauMax = order > 0 ? nodes[0]->t : t;
			narcs++;
		}
		else if(index + 1 == n)
		{
			tauMin = nodes[index - 1]->t;
			narcs++;
		}
		else
		{
			p = nodes[index]->pin;
			tauMin = nodes[index - 1]->t;
			tauMax = nodes[index]->t;

			if(nodes[index]->link != nodes[index - 1])
			{
				narcs++;
			}
		}

		dtau = tauMax - tauMin;
		ti = Uniform(tauMin, tauMax);
		tf = Exponential(ti, OMEGA_P);
		dt = tf - ti;

#ifdef DEBUG
		printf("ti = %lf, tf = %lf, dtau = %lf\n", ti, tf, dtau);
#endif

		if(tf > tauMax)
		{
			rejects++;
		}
		else
		{
			sigma = 1.0 / sqrt(tf - ti);
			q.x = Gaussian(sigma, 0.0);
			q.y = Gaussian(sigma, 0.0);
			q.z = Gaussian(sigma, 0.0);
			prob = Random();
			chance = pow(1.0 / sqrt(twopi * dt), 3.0) * vsqu * n * dtau * exp(dt * DotMultiplayVec3(&p, &q)) / (OMEGA_P * SquareVec3(&q) * narcs);

#ifdef DEBUG
			printf("|q| = %lf, R = %lf, Prob. = %lf\n", SquareVec3(&q), chance, prob);
#endif

			if(prob < chance)
			{
				InsertNodes(index, p, q, ti, tf);
			}
			else
			{
				rejects++;
			}
		}
	}
}

// Tries to change (extend if order > 0) the time
void ChangeTime()
{
	double tnew = Exponential(0.0, VectorEnergy(&k));

	if(order > 0)
	{
		tnew += nodes[2 * order - 1]->t;
	}

	if(tnew > TMAX)
	{
		rejects++;
	}
	else
	{
		t = tnew;
		tindex = (int)(t / dt);
	}
}

// Performs the update
void Update()
{
	double r = 4.0 * Random();

	if(r < 1.0)
	{
#ifdef DEBUG
		printf("Add ... \n");
#endif
		AddDiagram();
	}
	else if(r < 2.0)
	{
#ifdef DEBUG
		printf("Remove ... \n");
#endif
		RemoveDiagram();
	}
	else if(r < 3.0)
	{
#ifdef DEBUG
		printf("Swap ... \n");
#endif
		SwapDiagrams();
	}
	else
	{
#ifdef DEBUG
		printf("Change ... \n");
#endif
		ChangeTime();
	}
}

/*

	Tests

*/

// Tests the momentum conservation
int TestMomentumConservation(int order)
{
	int i;
	int n = 2 * order;
	vec3 test;
	double sum = 0.0;
	double expected = 0.0;
	printf("Momentum conservation ...\t");

	for(i = 0; i < n; i++)
	{
		double factor = 1.0;

		if(nodes[i]->t > nodes[i]->link->t)
		{
			factor = -1.0;
		}

		test.x = nodes[i]->pin.x - nodes[i]->pout.x - factor * nodes[i]->q.x;
		test.y = nodes[i]->pin.y - nodes[i]->pout.y - factor * nodes[i]->q.y;
		test.z = nodes[i]->pin.z - nodes[i]->pout.z - factor * nodes[i]->q.z;
		sum += SquareVec3(&test);
	}

	if(ABS(sum - expected) > EPSILON)
	{
		printf("\033[22;31mFAILURE\t[ %lf (%lf) ]\033[0m\n", sum, expected);
		return FALSE;
	}
	else
	{
		printf("\033[22;32mSUCCESS\t[ %lf (%lf) ]\033[0m\n", sum, expected);
		return TRUE;
	}
}

// Tests the diagram order (is the nodes array arranged correctly?)
int TestChronology(int order)
{
	int i;
	int n = 2 * order;
	double recent = 0.0;
	printf("Chronology of vertices ...\t");

	for(i = 0; i < n; i++)
	{
		if(nodes[i]->t < recent)
		{
			printf("\033[22;31mFAILURE\t[ %d (%lf .. %lf) ]\033[0m\n", i, recent, nodes[i]->t);
			return FALSE;
		}

		recent = nodes[i]->t;
	}

	if(recent > t)
	{
		printf("\033[22;31mFAILURE\t[ %d (%lf .. %lf) ]\033[0m\n", n, recent, t);
		return FALSE;
	}

	if(t > TMAX)
	{
		printf("\033[22;31mFAILURE\t[ t > TMAX (%lf) ]\033[0m\n", t);
		return FALSE;
	}
	
	printf("\033[22;32mSUCCESS\t[ %d OK ]\033[0m\n", n);
	return TRUE;
}

// Tests the diagram links (does every node have the right link?)
int TestLinks(int order)
{
	int i;
	int n = 2 * order;
	vec3 test;
	double sum = 0.0;
	double expected = 0.0;
	printf("Phonon propagator links ...\t");

	for(i = 0; i < n; i++)
	{
		node *mirror = nodes[i]->link;
		test.x = nodes[i]->q.x - mirror->q.x;
		test.y = nodes[i]->q.y - mirror->q.y;
		test.z = nodes[i]->q.z - mirror->q.z;
		sum += SquareVec3(&test);
	}

	if(ABS(sum - expected) > 0.0)
	{
		printf("\033[22;31mFAILURE\t[ %lf (%lf) ]\033[0m\n", sum, expected);
		return FALSE;
	}
	else
	{
		printf("\033[22;32mSUCCESS\t[ %lf (%lf) ]\033[0m\n", sum, expected);
		return TRUE;
	}
}

// Tests the current diagram state
void NodeTests()
{
	int state = 0;
	int total = 3;
	state += TestMomentumConservation(order);
	state += TestChronology(order);
	state += TestLinks(order);

	if(state != total)
	{
		printf("\033[22;31mCross check warning: \033[0m%d / %d tests failed\n", total - state, total);
	}
	else
	{
		printf("Cross check succeeded.\n");
	}
}

// Tests the uniform distribution
int TestUniform(int steps)
{
	int i;
	double total = 0.0;
	double min = 0.0;
	double max = 1.0;
	double mean = (max - min) / 2.0;
	printf("Uniform distribution ...\t");

	for(i = 0; i < steps; i++)
	{
		total += Uniform(min, max);
	}

	total = total / (double)steps;

	if(ABS(total - mean) > 0.01)
	{
		printf("\033[22;31mFAILURE\t[ %lf (%lf) ]\033[0m\n", total, mean);
		return FALSE;
	}
	else
	{
		printf("\033[22;32mSUCCESS\t[ %lf (%lf) ]\033[0m\n", total, mean);
		return TRUE;
	}
}

// Tests the discrete distribution
int TestDiscrete(int steps)
{
	int i;
	int min = 0;
	int max = 10;
	int value;
	double total = 0.0;
	double mean = (max - min) / 2.0;
	printf("Discrete distribution ...\t");

	for(i = 0; i < steps; i++)
	{
		value = Discrete(min, max + 1);

		if(value < min || value > max)
		{
			printf("\033[01;31mWARNING: OUT OF BOUNDS\t[ %d (%d : %d) ]\033[0m\n", value, min, max);
		}

		total += (double)value;
	}

	total = total / (double)steps;

	if(ABS(total - mean) > 0.01)
	{
		printf("\033[22;31mFAILURE\t[ %lf (%lf) ]\033[0m\n", total, mean);
		return FALSE;
	}
	else
	{
		printf("\033[22;32mSUCCESS\t[ %lf (%lf) ]\033[0m\n", total, mean);
		return TRUE;
	}
}

// Tests the normal distribution
int TestGaussian(int steps)
{
	int i;
	double total = 0.0;
	double mean = 0.0;
	printf("Gaussian distribution ...\t");

	for(i = 0; i < steps; i++)
	{
		total += Gaussian(1.0, mean);
	}

	total = total / (double)steps;

	if(ABS(total - mean) > 0.01)
	{
		printf("\033[22;31mFAILURE\t[ %lf (%lf) ]\033[0m\n", total, mean);
		return FALSE;
	}
	else
	{
		printf("\033[22;32mSUCCESS\t[ %lf (%lf) ]\033[0m\n", total, mean);
		return TRUE;
	}
}

// Tests the exponential distribution
int TestExponential(int steps)
{
	int i;
	double total = 0.0;
	double lambda = 1.0;
	double mean = 1.0 / lambda;
	printf("Exponential distribution ...\t");

	for(i = 0; i < steps; i++)
	{
		total += Exponential(0.0, lambda);
	}

	total = total / (double)steps;

	if(ABS(total - mean) > 0.01)
	{
		printf("\033[22;31mFAILURE\t[ %lf (%lf) ]\033[0m\n", total, mean);
		return FALSE;
	}
	else
	{
		printf("\033[22;32mSUCCESS\t[ %lf (%lf) ]\033[0m\n", total, mean);
		return TRUE;
	}
}

// Tests the Green0 function by integration
int TestIntegrate(int steps)
{
	int i;
	double total = 0.0;
	double t;
	double result = -1.0 / MU;
	printf("Zero-th order integral ...\t");

	for(i = 0; i < steps; i++)
	{
		t = Uniform(0, TMAX);
		total += exp(MU * t);
	}

	total = total / (double)steps;

	if(ABS(total - result) > 0.01)
	{
		printf("\033[22;31mFAILURE\t[ %lf (%lf) ]\033[0m\n", total, result);
		return FALSE;
	}
	else
	{
		printf("\033[22;32mSUCCESS\t[ %lf (%lf) ]\033[0m\n", total, result);
		return TRUE;
	}
}

// Tests the square of a vec3
int TestSquareVector()
{
	vec3 a;
	double result;
	double experiment;
	a.x = 5.3;
	a.y = 3.1;
	a.z = 0;
	result = 37.7;
	printf("Square of a three vector ...\t");
	experiment = SquareVec3(&a); 

	if(ABS(experiment - result) > EPSILON)
	{
		printf("\033[22;31mFAILURE\t[ %lf (%lf) ]\033[0m\n", experiment, result);
		return FALSE;
	}
	else
	{
		printf("\033[22;32mSUCCESS\t[ %lf (%lf) ]\033[0m\n", experiment, result);
		return TRUE;
	}
}

// Tests the addition of two vec3
int TestAddVector()
{
	vec3 a;
	vec3 b;
	vec3 c;
	double result;
	double experiment;
	a.x = 5.3;
	a.y = 3.1;
	a.z = 0;
	b.x = 1.3;
	b.y = 19.19;
	b.z = -5.2;
	printf("Add two three vectors ...\t");
	AddVec3(&a, &b, &c);
	result = 567.4441;
	experiment = SquareVec3(&c);

	if(ABS(experiment - result) > EPSILON)
	{
		printf("\033[22;31mFAILURE\t[ %lf (%lf) ]\033[0m\n", experiment, result);
		return FALSE;
	}
	else
	{
		printf("\033[22;32mSUCCESS\t[ %lf (%lf) ]\033[0m\n", experiment, result);
		return TRUE;
	}
}

// Tests the subtraction of two vec3
int TestSubtractVector()
{
	vec3 a;
	vec3 b;
	vec3 c;
	double result;
	double experiment;
	a.x = 5.3;
	a.y = 3.1;
	a.z = 0;
	b.x = 1.3;
	b.y = 19.19;
	b.z = -5.2;
	printf("Subtract two three vectors ...\t");
	SubtractVec3(&a, &b, &c);
	result = 301.9281;
	experiment = SquareVec3(&c);

	if(ABS(experiment - result) > EPSILON)
	{
		printf("\033[22;31mFAILURE\t[ %lf (%lf) ]\033[0m\n", experiment, result);
		return FALSE;
	}
	else
	{
		printf("\033[22;32mSUCCESS\t[ %lf (%lf) ]\033[0m\n", experiment, result);
		return TRUE;
	}
}

// Tests the product of two vec3
int TestProductVector()
{
	vec3 a;
	vec3 b;
	double result;
	double experiment;
	a.x = 5.3;
	a.y = 3.1;
	a.z = 0;
	b.x = 1.3;
	b.y = 19.19;
	b.z = -5.2;
	printf("Multiply vector with vector ...\t");
	result = 66.379;
	experiment = DotMultiplayVec3(&a, &b);

	if(ABS(experiment - result) > EPSILON)
	{
		printf("\033[22;31mFAILURE\t[ %lf (%lf) ]\033[0m\n", experiment, result);
		return FALSE;
	}
	else
	{
		printf("\033[22;32mSUCCESS\t[ %lf (%lf) ]\033[0m\n", experiment, result);
		return TRUE;
	}
}

// Tests the multiplication of a vec3 with a scalar
int TestMultiplyVector()
{
	vec3 a;
	vec3 c;
	double result;
	double experiment;
	a.x = 5.3;
	a.y = 3.1;
	a.z = 0;
	printf("Multiply vector with scalar ...\t");
	ScalarMultiplayVec3(-2.0, &a, &c);
	result = 150.8;
	experiment = SquareVec3(&c);

	if(ABS(experiment - result) > EPSILON)
	{
		printf("\033[22;31mFAILURE\t[ %lf (%lf) ]\033[0m\n", experiment, result);
		return FALSE;
	}
	else
	{
		printf("\033[22;32mSUCCESS\t[ %lf (%lf) ]\033[0m\n", experiment, result);
		return TRUE;
	}
}

// Performs some initial tests
void InitialTests()
{
	int state = 0;
	const int total = 10;
	const int steps = 10000000;
	state += TestUniform(steps);
	state += TestDiscrete(steps);
	state += TestGaussian(steps);
	state += TestExponential(steps);
	state += TestIntegrate(steps);
	state += TestSquareVector();
	state += TestSubtractVector();
	state += TestAddVector();
	state += TestMultiplyVector();
	state += TestProductVector();

	if(state != total)
	{
		printf("\033[22;31mSome tests failed\033[0m [ %d (Total: %d) ]\n", total - state, total);
		printf("Terminating program execution ...\n");
		exit(0);
	}

	printf("All tests have been executed successfully.\n");
}

/*

	The simulation

*/

// Performs the permanent loop for a certain duration
void MCLoop()
{
	ulong iter = 0;
	ulong test = 0;
	ulong chck = 0;
	ulong stat = 0;
	time_t start = time(NULL);

	while(TRUE)
	{
		++iter;
		Update();
		Measure();

		if(++test == TEST_FREQUENCY)
		{
			test = 0;
			NodeTests();
		}

		if(++stat == STATISTIC_FREQUENCY)
		{
			stat = 0;
			PrintStatistic(iter);
			SaveStatistic();
		}

		if(++chck == CHECK_TIME_FREQUENCY)
		{
			chck = 0;
			time_t end = time(NULL);

#ifdef DEBUG
			printf("Measuring time ...\n");
			printf("start = %d, end = %d\n", (int)start, (int)end);
#endif

			if(end - start >= RUNTIME)
			{
				printf("Runtime limit exceeded ... Stopping the simulation ...\n");
				break;
			}
		}
	}
}

// Initializes some variables
void InitVariables()
{
	rejects = 0;
	order = 0;
	Z = (int*)calloc(HIGHEST_ORDER + 1, sizeof(int));
	nodes = (node**)malloc(sizeof(node*) * 2 * HIGHEST_ORDER);
	t = 0;
	tindex = 0;
	dt = TMAX / (double)NBINS;
	G = (double*)calloc(NBINS, sizeof(double));
    norm = (1.0 - VectorPropagator(&k, TMAX)) / VectorEnergy(&k);
}

// Starts the program
int main()
{
	SetRandomSeed();
   	InitVariables();
	SetSpecificSeed(200);

#ifdef INITIAL_TESTS

	InitialTests();

#endif
	
	MCLoop();
	return 0;
}